"""
This module contains modules for ui compontents used in the visualisation
of simulation problems.
"""
import pygame
from pygame.surface import Surface
from dataclasses import dataclass, field
from typing import List

from simpn.visualisation.modules.base import ModuleInterface
from simpn.visualisation.base import TUE_GREY, TUE_BLUE, TUE_LIGHTBLUE, LINE_WIDTH, TEXT_SIZE, TUE_RED, LINE_WIDTH
from simpn.assets import get_img_asset
from simpn.visualisation.events import check_event, NODE_CLICKED, SELECTION_CLEAR
from simpn.visualisation.text import prevent_overflow_while_rendering
from simpn.simulator import Describable

class UIClockModule(ModuleInterface):
    """
    This modules handles creating a clock in the top right corner to
    show the current simulation time to a predefined precision.

    :param `precision=2`:
        The preicision of the clock to show in the UI.
    """

    CLOCK_SIZE = (50, 50)
    OFFSET = 16

    def __init__(self, precision:int=2):
        super().__init__()
        self._precision = max(1, precision)
        self._time = 0
        self._target = 0
        self._pusher = 1.0 / (self._precision + 1)
        self._font_size = 12

    def create(self, sim, *args, **kwargs):
        self._time = sim.clock
        self._format = f"{round(self._time, self._precision)}"
        self._clock_img = pygame.image.load(get_img_asset("clock.png"))
        self._clock_img = pygame.transform.smoothscale(
            self._clock_img, self.CLOCK_SIZE
        )
        self._clock_rect = self._clock_img.get_rect()
        self._font = pygame.font.SysFont('Calibri', self._font_size)

    def pre_event_loop(self, sim, *args, **kwargs):
        self._target = sim.clock
        if self._time < self._target:
            push = max(self._pusher, (self._target - self._time) * self._pusher)
            self._time = min(self._target, self._time + push)
        self._format = f"{round(self._time, self._precision)}"

    def render_ui(self, window:Surface, *args, **kwargs):
        self._clock_rect.center = (
            self.OFFSET + self._clock_rect.width // 2, 
            window.get_height() - self.OFFSET - self._font_size - self._clock_rect.height // 2
        )

        # draw holder for text
        clock_text_rect = pygame.Rect(
            self.OFFSET,
            window.get_height() - self.OFFSET - self._font_size,
            self._clock_rect.width,
            24
        )
        pygame.draw.rect(
            window, TUE_LIGHTBLUE,
            (clock_text_rect.x, clock_text_rect.y,
             clock_text_rect.width, clock_text_rect.height),
            border_radius=5
        )
        pygame.draw.rect(
            window, TUE_BLUE,
            (clock_text_rect.x, clock_text_rect.y,
             clock_text_rect.width, clock_text_rect.height),
            LINE_WIDTH,
            border_radius=5
        )
        
        # handle the font and show it on the scren
        font_length = len(self._format)
        show = self._format[:font_length]
        label = self._font.render(show, True, TUE_BLUE)
        while label.get_width() > clock_text_rect.width * 0.9:
            font_length -= 1
            show = self._format[:font_length]
            label = self._font.render(show, True, TUE_BLUE)

        text_x_pos = clock_text_rect.x \
            + (clock_text_rect.width//2) \
            - (label.get_width()// 2)
        text_y_pos = clock_text_rect.y \
            + (clock_text_rect.height//2) \
            - (label.get_height()//2)
        window.blit(label, (text_x_pos, text_y_pos))

        #blit image for clock
        window.blit(self._clock_img, self._clock_rect)

    def handle_event(self, event, *args, **kwargs):
        # check if the event is a mousebuttondown
        if event.type == pygame.MOUSEBUTTONDOWN:
            if self._clock_rect.collidepoint(event.pos):
                # on left click increase precision
                if event.button == 1:
                    self._precision += 1
                # on right click decrease precision
                elif event.button == 3:
                    self._precision = max(1, self._precision - 1)

        return True

class UISidePanelModule(ModuleInterface):
    """"
    This module is a revealable side panel that can be adjusted to
    show different information packets.

    For the moment it will only trigger when a modelled node
    is clicked or "selected".
    """

    PANEL_CLICKER_SIZE = (16,40)
    PANEL_SIZE = (300,0)
    PUSH_OUT_SPEED = 20
    PUSH_OUT_MAX = PANEL_SIZE[0] - 20
    PANEL_Y_POS = 0

    TITLE_FONT_SIZE = 18
    SUBTITLE_FONT_SIZE = 16
    FONT_SIZE = 14    

    def __init__(self):
        super().__init__()
        self._opened = False
        self._push_out = 0
        self._description = None # if the description is None, nothing is shown
        self._selected = None
        self.reset_and_hide_description()

    def reset_and_hide_description(self):
        self._selected = None
        self._description = None

    def pre_event_loop(self, sim, *args, **kwargs):
        if self._opened:
            self._push_out = min(self.PUSH_OUT_MAX, self._push_out + self.PUSH_OUT_SPEED)
        else:
            self._push_out = max(0, self._push_out - self.PUSH_OUT_SPEED)
        if self._selected is not None:
            self._description = self._selected._model_node.get_description()

    def create(self, sim, *args, **kwargs):
        self.panel = pygame.rect.Rect(
            0, self.PANEL_Y_POS, self.PANEL_SIZE[0] + self.PANEL_CLICKER_SIZE[0], self.PANEL_SIZE[1]
        )
        self.open_button = pygame.image.load(get_img_asset("flip_open.png"))
        self.open_button = pygame.transform.rotate(self.open_button, -90)
        self.open_button = pygame.transform.smoothscale(
            self.open_button, self.PANEL_CLICKER_SIZE
        )
        self.orect = self.open_button.get_rect()
        self.close_button = pygame.image.load(get_img_asset("flip_close.png"))
        self.close_button = pygame.transform.rotate(self.close_button, -90)
        self.close_button = pygame.transform.smoothscale(
            self.close_button, self.PANEL_CLICKER_SIZE
        )
        self.crect = self.close_button.get_rect()

        return super().create(sim, *args, **kwargs)
        
    def handle_event(self, event, *args, **kwargs):
        if event.type == pygame.MOUSEBUTTONDOWN:
            if self.panel.collidepoint(event.pos):
                if self.orect.collidepoint(event.pos) and not self._opened:
                    if event.button == 1:
                        self._opened = True
                        self._push_out = 0
                elif self.crect.collidepoint(event.pos) and self._opened:
                    if event.button == 1:
                        self._opened = False
                return False
            
        if check_event(event, NODE_CLICKED):
            self._selected = event.node
            self._description = self._selected._model_node.get_description()
        elif check_event(event, SELECTION_CLEAR):
            self.reset_and_hide_description()

        return True
    
    def render_ui(self, window, *args, **kwargs):
        right = window.get_width()
        mid = window.get_height() // 2
        top = 150

        self.orect.y = mid 
        self.orect.x = right - self.orect.width - self._push_out
        self.crect.y = mid 
        self.crect.x = right - self.crect.width - self._push_out
        self.panel.x = right - self._push_out - self.crect.width
        self.panel.height = window.get_height() - self.panel.y

        # if self._push_out > 0:
        pygame.draw.rect(
            window, TUE_LIGHTBLUE, self.panel,
            border_radius=15
        )
        pygame.draw.rect(
            window, TUE_BLUE, self.panel,
            LINE_WIDTH, border_radius=15
        )

        if (not self._opened):
            window.blit(self.open_button, self.orect)
        else:
            window.blit(self.close_button, self.crect)

        if self._push_out >= 0 and self._description is not None:
            self._render_description(window)

        return super().render_ui(window, *args, **kwargs)
    
    def _render_description(self, window:pygame.surface.Surface):
        info_surface = pygame.surface.Surface(self.panel.size, pygame.SRCALPHA)
        pos_x = 32
        pos_y = 16
        line_offset = 8
        frame_height = 65
        
        normal_font = pygame.font.SysFont('Calibri', self.FONT_SIZE)
        heading_font = pygame.font.SysFont('Calibri', self.TITLE_FONT_SIZE, bold=True)
        bold_font = pygame.font.SysFont('Calibri', self.FONT_SIZE, bold=True)
        
        for (text, style) in self._description:
            if style == Describable.Style.HEADING:
                font = heading_font
            elif style == Describable.Style.BOLD:
                font = bold_font
            else:
                font = normal_font                
            org_y = pos_y
            _, pos_y = prevent_overflow_while_rendering(
                info_surface,
                lambda t: font.render(t, True, TUE_BLUE),
                self.panel.width - 106,
                text,
                (pos_x, pos_y) if not style == Describable.Style.BOXED else (pos_x + line_offset/2, pos_y + line_offset/2),
                line_offset
            )
            if style == Describable.Style.BOXED:
                pygame.draw.rect(
                    info_surface, TUE_BLUE, pygame.Rect(pos_x, org_y + line_offset/4, self.panel.width - 96, pos_y - org_y),
                    LINE_WIDTH,
                    border_radius=5
                )
                pos_y += line_offset/2    
        window.blit(info_surface, self.panel)
        
        
if __name__ == "__main__":
    from simpn.simulator import SimProblem, SimToken
    from simpn.helpers import Place, Transition
    from simpn.visualisation import Visualisation

    from random import normalvariate

    problem = SimProblem()

    class Start(Place):
        model=problem
        name="start"
        amount=5

    class Resource(Place):
        model=problem
        name="resource"
        amount=1

    class Action(Transition):
        model=problem
        name="Task One"
        incoming = ["start", "resource"]
        outgoing = ["end", "resource"]

        def behaviour(c, r):
            return [
                SimToken(c, delay=normalvariate(2,0.25)),
                SimToken(r, delay=normalvariate(2,0.25))
            ]

    vis = Visualisation(
        problem,
        extra_modules=[
            UIClockModule(3),
            UISidePanelModule()
        ]
    )

    vis.show()



from .bpmn import BPMN
from .bpmn import HelperBPMNEnd, HelperBPMNExclusiveJoin, HelperBPMNStart
from .bpmn import HelperBPMNExclusiveSplit, HelperBPMNIntermediateEvent
from .bpmn import HelperResourcePool, HelperBPMNFlow
from .bpmn import HelperBPMNParallelJoin, HelperBPMNParallelSplit

from .petrinets import Place, Transition


"""
Module to handle the creation and handling of custom events
within the pygame loop.
"""

from dataclasses import dataclass, field
from copy import deepcopy
from typing import Dict
from pygame.event import Event
from pygame import USEREVENT 

### list of known custom events
## Node related
NODE_CLICKED = "node.clicked"
SELECTION_CLEAR = "selection.clear"
    

def create_event(type:str, context:Dict[str,object]) -> Event:
    """
    Creates a custom pygame event to throw into the event que.
    """
    context = context
    context['named_type'] = type
    return Event(
        USEREVENT + 1,
        context
    )

def check_event(event:Event, type:str) -> bool:
    """
    Checks whether the event is a custom event and has the desired named
    type.
    """
    if event.type == USEREVENT +1 :
        try:
            return event.named_type == type
        except AttributeError:
            return False
    return False


"""
This module describes a set of meta classes that allow for subclassing 
to create a simulation using the prototypes for a BPMN model.

To use these helper classes, import one of the desired prototype helpers,
or import `BPMN` to use the general class which will route to the desired
helper class for a prototype.
"""

from simpn.prototypes import *
from simpn.simulator import SimProblem

from typing import List, Literal, Union
from abc import abstractmethod

class HelperBPMNTask(BPMNTask):
    """
    A helper class to create a `BPMNTask` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields and functions. After the defintion is created a new 
    prototype instance will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `outgoing`:
        A list of `SimVar` to pull tokens from to trigger the task.
    :fieldname `outgoing`:
        A list of `SimVar` to depack the tokens returned by the behaviour
        function into.

    ^^^
    Required class functions
    ^^^
    :fieldname `behaviour`:
        A function determines how long each token is held by the BPMNTask.
    """
    model = None
    incoming = None
    outgoing = None
    name = None

    @staticmethod
    def __create__(cls, **kwargs):
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or outgoing is None or name is None:
            # Don't register the abstract base class
            if cls.__name__ == 'HelperBPMNTask':
                return
            raise ValueError("You must define static class variables: model, incoming, outgoing, and name in your HelperBPMNTask subclass.")
        # Fetch static/class methods
        behaviour = getattr(cls, 'behaviour', None)
        if behaviour is None or not callable(behaviour):
            raise NotImplementedError("You must implement a static/class method 'behaviour(*args)' in your HelperBPMNTask subclass.")
        guard = getattr(cls, 'guard', None)
        outgoing_behaviour = getattr(cls, 'outgoing_behaviour', None)
        # If not implemented, set to None
        if guard is not None and not callable(guard):
            guard = None
        if outgoing_behaviour is not None and not callable(outgoing_behaviour):
            outgoing_behaviour = None
        # Register the task with the model by instantiating the subclass
        HelperBPMNTask(model, incoming, outgoing, name, behaviour, guard=guard, outgoing_behavior=outgoing_behaviour)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Fetch static/class variables
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or outgoing is None or name is None:
            # Don't register the abstract base class
            if cls.__name__ == 'HelperBPMNTask':
                return
            raise ValueError("You must define static class variables: model, incoming, outgoing, and name in your HelperBPMNTask subclass.")
        # Fetch static/class methods
        behaviour = getattr(cls, 'behaviour', None)
        if behaviour is None or not callable(behaviour):
            raise NotImplementedError("You must implement a static/class method 'behaviour(*args)' in your HelperBPMNTask subclass.")
        guard = getattr(cls, 'guard', None)
        outgoing_behaviour = getattr(cls, 'outgoing_behaviour', None)
        # If not implemented, set to None
        if guard is not None and not callable(guard):
            guard = None
        if outgoing_behaviour is not None and not callable(outgoing_behaviour):
            outgoing_behaviour = None
        # Register the task with the model by instantiating the subclass
        cls(model, incoming, outgoing, name, behaviour, guard=guard, outgoing_behavior=outgoing_behaviour)

    @staticmethod
    @abstractmethod
    def behaviour(*args):
        pass

class HelperBPMNStart(BPMNStartEvent):
    """
    A helper class to create a `BPMNStartEvent` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields and functions. After the defintion is created a new 
    prototype instance will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `outgoing`:
        A list of `SimVar` that to put tokens into once the arrival time
        occurs.

    ^^^
    Required class functions
    ^^^
    :fieldname `interarrival_time`:
        A function that returns the delay between the next token arrival.
    """
    model = None
    outgoing = None
    name = None
    amount = None

    @staticmethod
    def __create__(cls, **kwargs):
        model = getattr(cls, 'model', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        amount = getattr(cls, 'amount', 1)
        behaviour = getattr(cls, 'behaviour', None)
        if model is None or outgoing is None or name is None:
            if cls.__name__ == 'HelperBPMNStart':
                return
            raise ValueError("You must define static class variables: model, outgoing, and name in your HelperBPMNStart subclass.")
        interarrival_time = getattr(cls, 'interarrival_time', None)
        if interarrival_time is None or not callable(interarrival_time):
            raise NotImplementedError("You must implement a static/class method 'interarrival_time()' in your HelperBPMNStart subclass.")
        # Register the start event with the model by instantiating BPMNStartEvent
        HelperBPMNStart(model, [], outgoing, name, interarrival_time, behaviour)

    def __init_subclass__(cls, **kwargs):
        model = getattr(cls, 'model', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        amount = getattr(cls, 'amount', 1)
        if model is None or outgoing is None or name is None:
            if cls.__name__ == 'HelperBPMNStart':
                return
            raise ValueError("You must define static class variables: model, outgoing, and name in your HelperBPMNStart subclass.")
        interarrival_time = getattr(cls, 'interarrival_time', None)
        if interarrival_time is None or not callable(interarrival_time):
            raise NotImplementedError("You must implement a static/class method 'interarrival_time()' in your HelperBPMNStart subclass.")
        # Register the start event with the model by instantiating BPMNStartEvent
        cls(model, [], outgoing, name, interarrival_time)

    @staticmethod
    @abstractmethod
    def interarrival_time():
        pass

class HelperBPMNEnd(BPMNEndEvent):
    """
    A helper class to create a `BPMNEndEvent` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields and functions. After the defintion is created a new 
    prototype instance will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of `SimVar` that require tokens to trigger the event.
    """
    model = None
    incoming = None
    name = None

    @staticmethod
    def __create__(cls, **kwargs):
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or name is None:
            if cls.__name__ == 'HelperBPMNEnd':
                return
            raise ValueError("You must define static class variables: model, outgoing, and name in your HelperBPMNStart subclass.")
        # Register the start event with the model by instantiating BPMNStartEvent
        HelperBPMNEnd(model, incoming, [], name)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or name is None:
            if cls.__name__ == 'HelperBPMNEnd':
                return
            raise ValueError("You must define static class variables: model, outgoing, and name in your HelperBPMNStart subclass.")
        # Register the start event with the model by instantiating BPMNStartEvent
        cls(model, incoming, [], name)

class HelperBPMNIntermediateEvent(BPMNIntermediateEvent):
    """
    A helper class to create a `BPMNIntermediateEvent` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields and functions. After the defintion is created a new 
    prototype instance will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of `SimVar` that require tokens to trigger the event.
    :fieldname `outgoing`:
        A list of `SimVar` to place the tokens into.

    ^^^
    Required class functions
    ^^^
    :fieldname `behaviour`:
        A function that takes tokens matching the length of `incoming` and
        returns a list of tokens to place into `outgoing` places.
    """
    model = None
    incoming = None
    outgoing = None
    name = None

    @staticmethod
    def __create__(cls, **kwargs):
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or outgoing is None or name is None:
            if cls.__name__ == 'HelperBPMNIntermediateEvent':
                return
            raise ValueError("You must define static class variables: model, incoming, outgoing, and name in your HelperBPMNIntermediateEvent subclass.")
        behaviour = getattr(cls, 'behaviour', None)
        if behaviour is None or not callable(behaviour):
            raise NotImplementedError("You must implement a static/class method 'behaviour(*args)' in your HelperBPMNIntermediateEvent subclass.")
        # Register the intermediate event with the model by instantiating BPMNIntermediateEvent
        HelperBPMNIntermediateEvent(model, incoming, outgoing, name, behaviour)


    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        model = getattr(cls, 'model', None)
        incoming = getattr(cls, 'incoming', None)
        outgoing = getattr(cls, 'outgoing', None)
        name = getattr(cls, 'name', None)
        if model is None or incoming is None or outgoing is None or name is None:
            if cls.__name__ == 'HelperBPMNIntermediateEvent':
                return
            raise ValueError("You must define static class variables: model, incoming, outgoing, and name in your HelperBPMNIntermediateEvent subclass.")
        behaviour = getattr(cls, 'behaviour', None)
        if behaviour is None or not callable(behaviour):
            raise NotImplementedError("You must implement a static/class method 'behaviour(*args)' in your HelperBPMNIntermediateEvent subclass.")
        # Register the intermediate event with the model by instantiating BPMNIntermediateEvent
        cls(model, incoming, outgoing, name, behaviour)

    @staticmethod
    @abstractmethod
    def behaviour(*args):
        pass

class HelperBPMNExclusiveSplit(BPMNExclusiveSplitGateway):
    """
    A helper class to create a `BPMNExclusiveSplitGateway` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of `SimVar` with a single place to pull tokens from.
    :fieldname `outgoing`:
        A list of possible `SimVar` to place the tokens into.

    ^^^
    Required class functions
    ^^^
    :fieldname `choice`:
        A function that takes a token from `incoming` and
        returns a list of the same length of `outgoing` where 
        some elements are `None` and some are tokens place into `outgoing` 
        places.
    """
    model = None
    incoming = None
    outgoing = None
    name = None

    @staticmethod
    def __create__(cls, **kwargs):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        if not hasattr(cls, "choice"):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define a 'choice' method.")
        # Register the gateway automatically
        HelperBPMNExclusiveSplit(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
            cls.choice,
            **kwargs
        )

    def __init_subclass__(cls):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        if not hasattr(cls, "choice"):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define a 'choice' method.")
        # Register the gateway automatically
        cls(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
            cls.choice
        )

    @staticmethod
    @abstractmethod
    def choice(c):
        pass

class HelperBPMNExclusiveJoin(BPMNExclusiveJoinGateway):
    """
    A helper class to create a `BPBPMNExclusiveJoinGatewayMN` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of possible `SimVar` to pull tokens from for joining.
    :fieldname `outgoing`:
        A list of `SimVar` with a single element to place the joined tokens into.
    """
    model = None
    incoming = None
    outgoing = None
    name = None

    @staticmethod
    def __create__(cls, **kwargs):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        # Register the gateway automatically
        HelperBPMNExclusiveJoin(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name
        )

    def __init_subclass__(cls):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        # Register the gateway automatically
        cls(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name
        )

class HelperResourcePool:
    """
    A helper subclass instance to make a resource pool within the simulation.
    Basically, adds a SimVar place with an `amount` of tokens for use. But, 
    uses a thin wrapper around SimVar for visualiusation of the place and edges.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `amount`:
        The amount of tokens to initially place in this resource pool. Tokens
        are generated using the name of the place plus an integer identifier
    """
    name:str=None 
    model:SimProblem=None 
    amount:int=None

    def __create__(cls, **kwargs):
        if any(hasattr(cls, attr) and getattr(cls, attr) is None for attr in ["name","model","amount"]):
            raise ValueError('Missing values for the following key attributes: ["name","model","amount"]')
        place = cls.model.add_place(cls.name)
        for i in range(cls.amount):
            place.put(f"{cls.name}-{i+1}") 
        place._resource_pool = True

    def __init_subclass__(cls, **kwargs):
        if not all(hasattr(cls, name) and getattr(cls, name) is not None for name in ["name","model","amount"]):
            raise ValueError('Missing values for the following key attributes: ["name","model","amount"]')
        place = cls.model.add_place(cls.name)
        for i in range(cls.amount):
            place.put(f"{cls.name}-{i+1}")
        place._resource_pool = True

class HelperBPMNFlow(BPMNFlow):
    """
    A helper class to create a `BPMNFlow` prototype. Which is a `SimVar`
    with a different visualisation

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    
    ^^^
    Optional class fields
    ^^^
    :fieldname `amount`:
    The amount of tokens to initially place in this resource pool. Tokens
    are generated using the name of the place plus an integer identifier
    """

    def __create__(cls, **kwargs):
        if any(hasattr(cls, attr) and getattr(cls, attr) is None for attr in ["name","model"]):
            raise ValueError('Missing values for the following key attributes: ["name","model"]')
        place = BPMNFlow(cls.model, cls.name)
        if hasattr(cls, 'amount'):
            for i in range(cls.amount):
                place.put(f"{cls.name}-{i+1}") 

    def __init_subclass__(cls, **kwargs):
        if not all(hasattr(cls, name) and getattr(cls, name) is not None for name in ["name","model"]):
            raise ValueError('Missing values for the following key attributes: ["name","model"]')
        place = BPMNFlow(cls.model, cls.name)
        if hasattr(cls, 'amount'):
            for i in range(cls.amount):
                place.put(f"{cls.name}-{i+1}") 

class HelperBPMNParallelSplit(BPMNParallelSplitGateway):
    """
    A helper class to create a `BPMNParallelSplitGateway` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of `SimVar` with a single place to pull tokens from.
    :fieldname `outgoing`:
        A list of  `SimVar` to place 'copied' tokens into.
    """

    @staticmethod
    def __create__(cls, **kwargs):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        BPMNParallelSplitGateway(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
            **kwargs
        )

    def __init_subclass__(cls):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        # Register the gateway automatically
        cls(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
        )

class HelperBPMNParallelJoin(BPMNParallelJoinGateway):
    """
    A helper class to create a `BPMNParallelJoinGateway` prototype.

    ---
    Usage
    ---
    Create a new class definition that inherits from this class with the
    following fields. After the defintion is created a new prototype instance
    will be recorded in the simulation problem.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    :fieldname `incoming`:
        A list of `SimVar` that each require one copy of a token to trigger
        the join.
    :fieldname `outgoing`:
        A list of a single `SimVar` to place the reocmbined token into.

    ^^^^
    Optional class functions
    ^^^^
    :fieldname `behaviour`:
        A function that takes the parallel tokens and returns a list of 
        `Simtoken` to place into outgoing. Used to collect data attributes
        from parallel actions back into a single token.
    """

    @staticmethod
    def __create__(cls, **kwargs):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        if hasattr(cls, 'behaviour'):
            behaviour = cls.behaviour
        else:
            behaviour = None
        # Register the gateway automatically
        BPMNParallelJoinGateway(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
            behaviour,
            **kwargs
        )

    def __init_subclass__(cls):
        if not all(hasattr(cls, attr) for attr in ("model", "incoming", "outgoing", "name")):
            raise AttributeError("HelperBPMNExclusiveSplitGateway subclasses must define model, incoming, outgoing, and name class variables.")
        if hasattr(cls, 'behaviour'):
            behaviour = cls.behaviour
        else:
            behaviour = None
        # Register the gateway automatically
        cls(
            cls.model,
            cls.incoming,
            cls.outgoing,
            cls.name,
            behaviour
        )

TYPES = {
    "start" : HelperBPMNStart.__create__,
    "end" : HelperBPMNEnd.__create__,
    "task" : HelperBPMNTask.__create__,
    "gat-ex-split" : HelperBPMNExclusiveSplit.__create__,
    "gat-ex-join" : HelperBPMNExclusiveJoin.__create__,
    "event" : HelperBPMNIntermediateEvent.__create__,
    "resource-pool" : HelperResourcePool.__create__,
    "flow" : HelperBPMNFlow.__create__,
    "gat-para-split" : HelperBPMNParallelSplit.__create__,
    "gat-para-join" : HelperBPMNParallelJoin.__create__

}
TYPE_NAMES = Literal["start", "end", "task", "gat-ex-split",
                     "gat-ex-join", "event", "resource-pool",
                     "flow", "gat-para-split", "gat-para-join"]

class BPMN:
    """
    A synatical sugar wrapper class to handle making a specific bpmn helper.
    To use this class, create a class definition that inherits from this 
    class and include the approciate fields to make a new prototype in the
    given `SimProblem`.

    ^^^^
    Required class fields
    ^^^^
    :fieldname `type`: 
        a string to denote what prototype should be made.
    :expected values: 
        `start`, `end`, `task`, `gat-ex-split`, `gat-ex-join`, `event`, 
        `resource-pool`, `flow`, `gat-para-split`, `gat-para-join`
    :fieldname `model`:
        The `SimProblem` to add the prototype to.
    :fieldname `name`:
        A unique identifier for the new prototype. 
    ^^^
    Optional class fields
    ^^^
    :fieldname `incoming`:
        A list of strings or `SimVar`'s to use as incoming. If a string is 
        passed then that string will be used to either make a `SimVar` or
        look one up in the given model.
    :fieldname `outgoing`:
        A list of strings or `SimVar`'s to use as outoing. Similar to above.
    
    For the fields needed for each individual helper, please check the 
    helper class definition.

    ^^^
    Examples of usage
    ^^^
    .. code-block:: python
        class Start(BPMN):
            type="start"
            model=agency
            name="arrive"
            amount=1
            outgoing=["waiting"]

            def interarrival_time():
                return exp(1/10)
            
        class Scan(BPMN):
            type="task"
            model=agency
            name="scan"
            incoming=["waiting", "employee"]
            outgoing=["done", "employee"]

            def behaviour(c, r):
                return [SimToken((c, r), delay=exp(1/9))]
            
        class End(BPMN):
            type="end"
            model=agency
            name="complete"
            incoming=["done"]
    """
    
    type:TYPE_NAMES=None 
    model:SimProblem=None
    name:str=None
    incoming:List[Union[str,SimVar]]=None
    outgoing:List[Union[str,SimVar]]=None
    guard=None 

    def __init_subclass__(cls, **kwargs):
        if not all(hasattr(cls, attr) and getattr(cls, attr) is not None for attr in ("type", "model", "name")):
            raise ValueError("Missing required attributes for BPMN construct : ['model', 'name', 'type']")
        # handle outgoing strings
        if hasattr(cls, 'outgoing') and cls.outgoing != None:
            for i,val in enumerate(cls.outgoing):
                if isinstance(val, str):
                    if val in cls.model.id2node.keys():
                        cls.outgoing[i] = cls.model.id2node[val]
                    else:
                        cls.outgoing[i] = BPMNFlow(cls.model, val)# cls.model.add_var(val)
        # handle incoming strings
        if hasattr(cls, 'incoming') and cls.incoming != None:
            for i,val in enumerate(cls.incoming):
                if isinstance(val, str):
                    if val in cls.model.id2node.keys():
                        cls.incoming[i] = cls.model.id2node[val]
                    else:
                        cls.incoming[i] = BPMNFlow(cls.model, val)
        tasker = TYPES[cls.type]
        tasker(cls)




def sed(s1: str, s2: str) -> int:
    """
    Computes the string edit distance between two strings.

    :param s1: The first string.
    :param s2: The second string.

    :return: The edit distance between the two strings.
    """
    if len(s1) < len(s2):
        return sed(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def seds(s1: str, s2: str) -> float:
    """
    Computes the string edit distance similarity between two strings.
    This is the string edit distance divided by the length of the longer string.

    :param s1: The first string.
    :param s2: The second string.

    :return: The edit distance between the two strings.
    """
    return 1.0 - (sed(s1, s2) / max(len(s1), len(s2)))


"""
This module containers helpers for testing out the module pipeline 
through the `Visualisation` class.
"""

from .base import ModuleInterface

class CheckerModule(ModuleInterface):

    def create(self, sim, *args, **kwargs):
        self.create_called = True 
        self.pre_called = False 
        self.handle_event_called = False
        self.render_sim_called = False 
        self.render_ui_called = False 
        self.firing_called = False 
        self.post_called = False 

    def pre_event_loop(self, sim, *args, **kwargs):
        # print("[Checking] pre_event_loop called...")
        self.pre_called = True

    def handle_event(self, event, *args, **kwargs):
        self.handle_event_called = True 

    def render_sim(self, screen, *args, **kwargs):
        self.render_sim_called = True 

    def render_ui(self, window, *args, **kwargs):
        self.render_ui_called = True 

    def firing(self, fired, sim, *args, **kwargs):
        self.firing_called = True

    def post_event_loop(self, sim, *args, **kwargs):
        self.post_called = True



"""
A module which provide useful text related rendering helpers.
"""

from typing import Tuple
from pygame.surface import Surface

def prevent_overflow_while_rendering(
        surface:Surface,
        fonter:callable,
        width:float,
        text:str,
        initial_pos:Tuple[float,float],
        line_offset:float,
    ) -> Tuple[float, float]:
    """
    Helper function to ensure that the width of a text block
    does not overflow a given width. Returns the final position
    after preventing overflowing.
    """
    pos_x, pos_y = initial_pos

    while len(text) > 0:
        curr = 1
        rendered_text = fonter(text[:curr])
        while rendered_text.get_width() < width:
            curr += 1
            rendered_text = fonter(text[:curr])
            if (curr >= len(text)):
                break 

        text=text[curr:]
        surface.blit(rendered_text, (pos_x, pos_y))
        pos_y += rendered_text.get_height()
        if len(text) > 0 :
            pos_y += line_offset
    
    return (pos_x, pos_y)





"""
This module contains the glue and blueprints for making a visualisation
module to add new features to a visualisation on demand. The modules are
instantiated and passed to the `Visualisation` class from 
`simpn.visualisation`, where the class will pass along information and 
allow the visualisation to be altered at predefined hook locations.
"""

from pygame.event import Event
from pygame.surface import Surface

class ModuleInterface():
    """
    Ideally, modules are mostly stateless and rely on message passing 
    via the ``handle_event`` hook to consider if they need to do anything at 
    all.

    ## Interface
    Modules can implement the following functions as hooks into the
    visualisation pipeline to implement their behaviour\:

    :create(...): 
        called once within ``Visualisation.__init__``
    :pre_event_loop(...): 
        called once per cycle of game loop
    :handle_event(event,...): 
        called once for each event of a game loop
    :render_sim(screen,...):
        called once for each draw of a game loop, for drawing simulation
        elements that will be zoomable.
    :render_ui(window,...):
        called once for each draw of a game loop, for drawing ui elements
        that are drawn over the simulation and are not zoomable.
    :firing(fired,sim,...):
        called once in the play loop when the next step of the simulation
        is decided. Useful for creating diagnostic ui elements.
    :post_event_loop(sim,...): 
        called once per cycle of game loop
    """

    def __init__(self):
        pass 

    def create(self, sim, *args, **kwargs):
        """
        This hook is called once during the setup of the visualisation.
        This function should setup any internal pygame assets required or
        other functionalities at this point.
        """
        pass 

    def pre_event_loop(self, sim, *args, **kwargs):
        """
        This hook is called at the start of each game loop for the 
        visualisation.
        """
        pass 

    def handle_event(self, event:Event, *args, **kwargs) -> bool:
        """
        This hook is called once for each event produced in the game loop
        of the visualisation. Returns a truthy value to say that the event
        should propagate to other modules and the base visualisation.

        :param event: 
            the event to consider
        :return bool:
            whether the event should be propagated.
        """
        return True

    def render_sim(self, screen:Surface, *args, **kwargs):
        """
        This hook is called once in the draw loop and will be passed a 
        surface to draw onto. This function should be implemented to add
        any drawable elements into the simulation pane.
        """
        pass

    def render_ui(self, window:Surface, *args, **kwargs):
        """
        This hook is called once in the draw loop and will be passed a 
        surface to draw onto. This function should be implemented to add
        any drawable elements into the ui pane.
        """
        pass 

    def firing(self, fired, sim, *args, **kwargs):
        """
        This hook is called once per simulation step and is passed along
        information about the step of the simulation.
        """
        pass

    def post_event_loop(self, sim, *args, **kwargs):
        """
        This hook is called at the end of each game loop for the 
        visualisation.
        """
        pass



